#include <allegro.h>
#include <stdio.h>
#include <semaphore.h>

#include "grafik.h"
#include "fileio.h"
#include "menu.h"

static int loadimages(void);

static BITMAP *tempsprite;
static BITMAP *sprite[32];
static BITMAP *backs[32];
static BITMAP *menu_item[32];
static BITMAP *msg[32];
static char *printtext;
static char *printtime;
static char *spruchtext;

static int mb, screen_x, screen_y;

// Double Buffer
static BITMAP   *doubleBuffer = NULL;
static BITMAP   *cleanscreen  = NULL;
static BITMAP   *screenBuffer  = NULL;

static BITMAP *buffer = NULL;
static BITMAP *stretch_buffer = NULL;
static BITMAP *back = NULL;
static BITMAP *screensaver = NULL;

static sem_t screen_sem;


int init_graphic(void)
{
	return sem_init(&screen_sem, 0, 1);
}

int set_graphic_mode(int x, int y, int screenmode)
{
	int retval;
	
	sem_wait(&screen_sem);
	if(screenmode==_FULLSCREEN_)
	{
		set_color_depth(32);
		if(set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(24);
		if(set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(16);
		if(set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(15);
		if(set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
	}
	else
	{
		set_color_depth(32);
		if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(24);
		if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(16);
		if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
		set_color_depth(15);
		if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, x, y, 0, 0)>=0)
		{
			retval=loadimages();
			sem_post(&screen_sem);
			return retval;
		}
	}
	sem_post(&screen_sem);
	return 1;
}

int loadimages(void) {

	int x, returncode;
	char bildname[512];

	//Unter Windows gibt es wegen trans_blitting einen fehler, auerdem ist das ganze
	//viel langsamer, daher habe ich alle video_bitmaps wieder zu system_bitmaps gendert
	
	if(doubleBuffer==NULL)
	{
		set_color_conversion(COLORCONV_TOTAL);
		
		returncode=0;
		doubleBuffer = create_system_bitmap(SCREEN_W, SCREEN_H);
		cleanscreen = create_system_bitmap(SCREEN_W, SCREEN_H);
		back = create_system_bitmap(SCREEN_W, SCREEN_H);
		screenBuffer = create_system_bitmap(SCREEN_W, SCREEN_H);
		if(!doubleBuffer || !cleanscreen || !back || !screenBuffer)
			return 3;
		
		clear_bitmap(doubleBuffer);
		clear_bitmap(cleanscreen);
		clear_bitmap(screenBuffer);
		clear_bitmap(back);
		
		//alles als frei markieren
		for(x=0;x<32;x++)
		{
			msg[x]=NULL;
			backs[x]=NULL;
			sprite[x]=NULL;
			menu_item[x]=NULL;
		}
		for(x=0;x<3;x++)
		{
			msg[x]=NULL;
			sprintf(bildname,"data/msg%d.tga" ,x);
			if((tempsprite=load_bitmap(bildname, NULL))==NULL)
			{
				printf("Load error: %s\n", bildname);
				return 2;
			}
			msg[x]=create_system_bitmap(tempsprite->w, tempsprite->h);
			clear_bitmap(msg[x]);
			blit(tempsprite, msg[x], 0, 0, 0, 0, tempsprite->w, tempsprite->h);
			destroy_bitmap(tempsprite);
		}
		for(x=1;x<11;x++)
		{
			sprite[x]=NULL;
			sprintf(bildname,"data/Bild%d.tga" ,x);
			if((tempsprite=load_bitmap(bildname, NULL))==NULL)
			{
				printf("Load error: %s\n", bildname);
				return 2;
			}
			/*sprite[x]=create_video_bitmap(tempsprite->w, tempsprite->h);
			if(!sprite[x])*/
			sprite[x]=create_system_bitmap(tempsprite->w, tempsprite->h);
			clear_bitmap(sprite[x]);
			blit(tempsprite, sprite[x], 0, 0, 0, 0, tempsprite->w, tempsprite->h);
			destroy_bitmap(tempsprite);
		}
		for(x=0;x<9;x++)
		{
			menu_item[x]=NULL;
			sprintf(bildname,"data/menu/menu%d.tga" ,x);
			if((tempsprite=load_bitmap(bildname, NULL))==0)
			{
				printf("Load error: %s\n", bildname);
				return 2;
			}
			menu_item[x]=create_system_bitmap(tempsprite->w, tempsprite->h);
			clear_bitmap(menu_item[x]);
			blit(tempsprite, menu_item[x], 0, 0, 0, 0, tempsprite->w, tempsprite->h);
			destroy_bitmap(tempsprite);
		}
		for(x=0;x<32;x++)
		{
			backs[x]=NULL;
			sprintf(bildname,"data/back/back%d.tga" ,x);
			if((tempsprite=load_bitmap(bildname, NULL))==NULL)
			{
				if(x>0)
					break;		//anzahl backgrounds dynamisch erkennen, aber mindestens einer muss da sein
				printf("Load error: %s\n", bildname);
				return 2;
			}
			backs[x]=create_system_bitmap(SCREEN_W, SCREEN_H);
			clear_bitmap(backs[x]);
			stretch_blit(tempsprite, backs[x], 0 ,0, tempsprite->w, tempsprite->h, 0, 0, back->w, back->h);
			destroy_bitmap(tempsprite);
		}
		blit(backs[0], back, 0, 0, 0, 0, backs[0]->w, backs[0]->h);
		mb = 0;
		clear_to_color(screen, makecol(0,0,0));
		return returncode;
	}
	else
		return 0;
}

void set_screen_pos(int x, int y)
{
	screen_x=x;
	screen_y=y;
	return;	
}
int get_screen_x_pos(void)
{
	return screen_x;
}
int get_screen_y_pos(void)
{
	return screen_y;
}

int prepare_menu(int menunum, int item2_enabled)
{
	if(clear_menu(back))
		return 1;
	if(menunum==_MAIN_MENU_)
	{
		if(add_item(menu_item[0], menu_item[1], menu_item[2], _MARKED_))
			return 1;
		if(add_item(menu_item[3], menu_item[4], menu_item[5], item2_enabled))
			return 1;
		if(add_item(menu_item[6], menu_item[7], menu_item[8], _ENABLED_))
			return 1;
	}
	buffer = create_bitmap(SCREEN_W, SCREEN_H);
	draw_menu(buffer, 8);

	return 0;
}

int update_jumper_menu(void)
{
	update_back(back);
	return draw_menu(doubleBuffer, 8);
}

int unload_menu_blender(void)
{
	destroy_bitmap(buffer);
	return 0;
}

void save_screen(void)
{
	if(screensaver==NULL)
		screensaver = create_bitmap(doubleBuffer->w, doubleBuffer->h);
	blit(doubleBuffer, screensaver, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
	return;
}

void restore_screen(void)
{
	if(screensaver!=NULL)
		blit(screensaver, doubleBuffer, 0, 0, 0, 0, screensaver->w, screensaver->h);
	return;
}

int init_message_blender(int message)
{
	buffer = create_bitmap(SCREEN_W, SCREEN_H);
	acquire_bitmap(buffer);
	acquire_bitmap(doubleBuffer);
	blit(doubleBuffer, buffer, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
	release_bitmap(doubleBuffer);
	draw_sprite(buffer, msg[message], (SCREEN_W-msg[message]->w)/2, (SCREEN_H-msg[message]->h)/2); 
	release_bitmap(buffer);
	return 0;
}

int blend_all(int a, int blender_mode)
{
	acquire_bitmap(doubleBuffer);
	acquire_bitmap(buffer);
	
	if(blender_mode==_TRANS_)
		set_trans_blender(0, 0, 0, a);
	if(blender_mode==_DISSOLVE_)
		set_dissolve_blender(0, 0, 0, a);
	if(blender_mode==_BURN_)
		set_burn_blender(0, 0, 0, a);
	if(blender_mode==_NONE_)
		draw_sprite(doubleBuffer, buffer, 0, 0);
	else
		draw_trans_sprite(doubleBuffer, buffer, 0, 0);
	
	release_bitmap(buffer);
	release_bitmap(doubleBuffer);
	return 0;
}

int unload_message_blender(void)
{
	destroy_bitmap(buffer);
	return 0;
}

int init_spruch_blender(char * spruch)
{
	int w=text_length(font, spruch), h=text_height(font);
	spruchtext=spruch;
	buffer = create_system_bitmap(SCREEN_W, SCREEN_H);
	stretch_buffer = create_system_bitmap(w, h);
	acquire_bitmap(stretch_buffer);
	clear_to_color(stretch_buffer, makecol(255,0,255));
	textout_ex(stretch_buffer, font, spruchtext, 0, 0, makecol(255,255,64), -1);
	release_bitmap(stretch_buffer);
	acquire_bitmap(buffer);
	clear_bitmap(buffer);
	acquire_bitmap(doubleBuffer);
	blit(doubleBuffer, buffer, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
	release_bitmap(doubleBuffer);
	release_bitmap(buffer);
	return 0;
}

int blend_spruch(int a)
{
	int w=(double)stretch_buffer->w*1.25;
	int h=(double)stretch_buffer->h*1.25;
	acquire_bitmap(doubleBuffer);
	acquire_bitmap(buffer);
	blit(buffer, doubleBuffer, 0, 0, 0, 0, buffer->w, buffer->h);
	stretch_sprite(doubleBuffer, stretch_buffer, (SCREEN_W-w)/2, a, w, h);
	release_bitmap(buffer);
	release_bitmap(doubleBuffer);
	return 0;
}

int unload_spruch_blender(void)
{
	destroy_bitmap(buffer);
	destroy_bitmap(stretch_buffer);
	return 0;
}

int init_level_blender(int * leveldata) {

	int x,y,counter;
	
	load_clean_level(leveldata);

	buffer = create_bitmap(SCREEN_W, SCREEN_H);
	acquire_bitmap(buffer);
	clear_bitmap(buffer);
	acquire_bitmap(back);
	blit(back, buffer, 0, 0, 0, 0, back->w, back->h);
	release_bitmap(back);

	x=0;
	y=-1;
	// Das level ist 1160 Objekte gro, danach kommen 8 zahlen als Prfsumme usw., die nicht ausgegeben
	// werden mssen
	for(counter=0;counter<1160;counter++){
		
		if((counter%40)==0)
			y++;
		x=(counter%40);
		
		if(*(leveldata+counter+1)==1)
			draw_sprite(buffer, sprite[1],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==2)
			draw_sprite(buffer, sprite[2],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==3)
			draw_sprite(buffer, sprite[3],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==4)
			draw_sprite(buffer, sprite[4],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==5)
			draw_sprite(buffer, sprite[5],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==6)
			draw_sprite(buffer, sprite[7],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==7)
			draw_sprite(buffer, sprite[7],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==8)
			draw_sprite(buffer, sprite[8],screen_x+x*16,screen_y+y*16);
	}
	textout_centre_ex(buffer, font, printtext, SCREEN_W/2, screen_y+480, makecol(0,255,0), -1);
	textout_centre_ex(buffer, font, printtime, SCREEN_W/2, screen_y-16, makecol(0,255,0), -1);
	release_bitmap(buffer);
	return 0;
}

int unload_level_blender(void)
{
	destroy_bitmap(buffer);
	return 0;
}


int load_clean_level(int * leveldata) {

	int x,y,counter;
	
	acquire_bitmap(cleanscreen);	
	acquire_bitmap(back);
	blit(back, cleanscreen, 0, 0, 0, 0, back->w, back->h);
	release_bitmap(back);
	
	x=0;
	y=-1;
	// Das level ist 1160 Objekte gro, danach kommen 8 zahlen als Prfsumme usw., die nicht ausgegeben
	// werden mssen
	for(counter=0;counter<1160;counter++){
		
		if((counter%40)==0)
			y++;
		x=(counter%40);
		
		if(*(leveldata+counter+1)==1)
			draw_sprite(cleanscreen, sprite[1],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==2)
			draw_sprite(cleanscreen, sprite[2],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==3)
			draw_sprite(cleanscreen, sprite[3],screen_x+x*16,screen_y+y*16);
		else if(*(leveldata+counter+1)==8)
			draw_sprite(cleanscreen, sprite[8],screen_x+x*16,screen_y+y*16);
	}
	release_bitmap(cleanscreen);
	return 0;
}

int add_remote_clean_level(int * leveldata) {

	int x,y,counter;
	
	acquire_bitmap(cleanscreen);	
	x=0;
	y=-1;
	// Das level ist 1160 Objekte gro, danach kommen 8 zahlen als Prfsumme usw., die nicht ausgegeben
	// werden mssen
	for(counter=0;counter<1160;counter++){
		
		if((counter%40)==0)
			y++;
		x=(counter%40);
		
		set_trans_blender(0, 0, 0, 128);
		if(*(leveldata+counter+1)==8)
			draw_trans_sprite(cleanscreen, sprite[10],screen_x+x*16,screen_y+y*16);
	}
	release_bitmap(cleanscreen);
	return 0;
}

int load_new_back(int x)
{
	if(backs[x]==NULL)
		return 1;
	acquire_bitmap(backs[x]);
	acquire_bitmap(back);
	blit(backs[x], back, 0, 0, 0, 0, backs[x]->w, backs[x]->h);
	release_bitmap(back);
	release_bitmap(backs[x]);
	return 0;
}

int move(int xsprite, int x, int y) {

	acquire_bitmap(doubleBuffer);
	if(xsprite==_JUMPER_)
	{
		blit(cleanscreen, doubleBuffer, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
		draw_sprite(doubleBuffer, sprite[7],screen_x+x,screen_y+y);
	}
	else if(xsprite==_JUMPER1_TRANS_)
	{
		set_trans_blender(0, 0, 0, 255);
		draw_trans_sprite(doubleBuffer, sprite[7], screen_x+x,screen_y+y); 
	}
	else if(xsprite==_JUMPER2_TRANS_)
	{
		blit(cleanscreen, doubleBuffer, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
		set_trans_blender(0, 0, 0, 128);
		draw_trans_sprite(doubleBuffer, sprite[9], screen_x+x,screen_y+y); 
	}
	else if(xsprite==_PACMAN_)
	{
		if(mb>=255)
			mb=0;
		rotate_sprite(doubleBuffer, sprite[5], screen_x+x, screen_y+y, itofix(mb));
		mb+=4;
	}
	release_bitmap(doubleBuffer);	
	return 0;
}

void print_status(char *statustext)
{
	acquire_bitmap(doubleBuffer);
	textout_centre_ex(doubleBuffer, font, printtext, SCREEN_W/2, screen_y+480, makecol(0,255,0), -1);
	release_bitmap(doubleBuffer);	
	printtext=statustext;
	return;
}

void set_status(char *statustext)
{
	printtext=statustext;
	return;
}

void print_time(char *timetext)
{
	acquire_bitmap(doubleBuffer);
	textout_centre_ex(doubleBuffer, font, timetext, SCREEN_W/2, screen_y-16, makecol(0,255,0), -1);
	release_bitmap(doubleBuffer);	
	printtime=timetext;
	return;
}

void set_time(char *timetext)
{
	printtime=timetext;
	return;
}

void reload_doublebuffer(void)
{
	acquire_bitmap(screenBuffer);
	acquire_bitmap(doubleBuffer);
	blit(screenBuffer, doubleBuffer, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
	release_bitmap(doubleBuffer);	
	release_bitmap(screenBuffer);
	return;
}

void show(void)
{
	sem_wait(&screen_sem);
	acquire_screen();
	acquire_bitmap(doubleBuffer);
	acquire_bitmap(screenBuffer);
	blit(doubleBuffer, screen, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
	blit(doubleBuffer, screenBuffer, 0, 0, 0, 0, doubleBuffer->w, doubleBuffer->h);
	release_bitmap(doubleBuffer);
	release_bitmap(screenBuffer);	
	release_screen();
	sem_post(&screen_sem);
	return;
}

void blend_game_over(int i)
{
	acquire_bitmap(doubleBuffer);
	stretch_blit(doubleBuffer, cleanscreen, 0, 0, doubleBuffer->w, doubleBuffer->h, i/4, i/2, cleanscreen->w-i/2, cleanscreen->h-i);
	blit(cleanscreen, doubleBuffer, 0, 0, 0, 0, cleanscreen->w, cleanscreen->h);
	release_bitmap(doubleBuffer);
	return;
}
