#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#ifdef _WIN32
	//Headerfiles fr Windows
	#include <windows.h>
	#include <winsock.h>
	#include <io.h>
#else
	//Headerfiles fr UNIX/Linux
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <netdb.h>
	#include <arpa/inet.h>
	#include <unistd.h>
#endif
#include "net.h"


#define PORT 1234
#ifdef _WIN32
	SOCKET global_socket=INVALID_SOCKET;
	SOCKET global_listen=INVALID_SOCKET;
#else
	int global_socket=-1;
	int global_listen=-1;
#endif
struct sockaddr_in global_server;
static int _protocol_send=0;


int create_server(void)
{
	struct sockaddr_in server;
	#ifdef _WIN32
		SOCKET sock;
	#else
		int sock;
	#endif
	
	#ifdef _WIN32
		//Initialisiere TCP fr Windows ("winsock")
		WORD wVersionRequested;
		WSADATA wsaData;
		wVersionRequested = MAKEWORD (1, 1);
		if(WSAStartup(wVersionRequested, &wsaData) != 0)
			return -1;
	#endif
	
	//Erzeuge die Socketadresse des Servers
	memset(&server, 0, sizeof (server));
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = htonl(INADDR_ANY);
	server.sin_port = htons(PORT);
	
	//Erzeuge den Socket
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock<0)
		return -1;
	
	if(bind(sock,(struct sockaddr*)&server, sizeof(server)) < 0)
		return -1;
	
	if(listen(sock, 1)==-1)
		return -1;
	
	memset(&global_server, 0, sizeof(server));
	global_listen=sock;
	return 0;
}

int poll_server_connection(void)
{
	char msg_buffer[1024];
	int msg_num=_MSG_CONNECT_, protocol_version=_PROTOCOL_VERSION_(0,2);
	struct timeval zeit;
	int ret_val;
	#ifdef _WIN32
		FD_SET empfang;
	#else
		fd_set empfang;
	#endif
	FD_ZERO(&empfang);
	
	//listening socket hinzufgen
	FD_SET(global_listen, &empfang);
	//kein timeout (sofort zurckkehren)
	zeit.tv_sec=0;
	zeit.tv_usec=0;
	ret_val=select(global_listen+1, &empfang, NULL, NULL, &zeit);
	if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
		return 0;
	if(FD_ISSET(global_listen, &empfang))
	{
		if((global_socket=accept(global_listen, NULL, NULL))<0)
			return -1;
		else
		{
			//erstmal version senden
			memcpy(msg_buffer, &msg_num, sizeof(msg_num));
			memcpy(msg_buffer+sizeof(msg_num), &protocol_version, sizeof(protocol_version));
			if(send(global_socket, msg_buffer, sizeof(msg_num)+sizeof(protocol_version), 0)<0)
				return -1;
			else
				return 1;
		}
	}
	
	return 0;
}

int wait_for_connect(void)
{
	char message[1024];
	int size, ret_val;
	int msg_num=_MSG_CONNECT_, protocol_version=_PROTOCOL_VERSION_(0,2);
	struct timeval zeit;
	
	memset(message, 0, 1024);

	#ifdef _WIN32
		FD_SET empfang;
		if(global_listen==INVALID_SOCKET)
			return -1;
	#else	
		fd_set empfang;
		if(global_listen==-1)
			return -1;
	#endif
	
	ret_val=poll_server_connection();
	if(ret_val==-1)
		return -1;
	else if(!ret_val)
		return 0;
	
	FD_ZERO(&empfang);
	FD_SET(global_socket, &empfang);
	zeit.tv_sec=0;
	zeit.tv_usec=0;
	ret_val=select(global_socket+1, &empfang, NULL, NULL, &zeit);
	if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
		return -1;
	if(FD_ISSET(global_socket, &empfang))
	{
		size=recv(global_socket, message, 1024, 0);
		if(size<1)
		{
			close_network();
			return -2;
		}
		memcpy(&msg_num, message, sizeof(msg_num));
		memcpy(&protocol_version, message+sizeof(msg_num), sizeof(protocol_version));
		if((msg_num==_MSG_CONNECT_) && (protocol_version==_PROTOCOL_VERSION_(0,2)))
		{
			#ifdef _WIN32
				closesocket(global_listen);
				global_listen=INVALID_SOCKET;
			#else
				close(global_listen);
				global_listen=-1;
			#endif
			return 1;
		}
		else
			return -2;
	}
	return 0;
}

int create_client(char *host)
{
	struct sockaddr_in server;
	struct hostent *host_info;
	#ifdef _WIN32
		SOCKET sock;
	#else
		int sock;
		long new_fd, old_fd;
	#endif
	
	#ifdef _WIN32
		//Initialisiere TCP fr Windows ("winsock")
		WORD wVersionRequested;
		WSADATA wsaData;
		wVersionRequested = MAKEWORD (1, 1);
		if(WSAStartup (wVersionRequested, &wsaData) != 0)
			return -1;
	#endif
	
	host_info = gethostbyname(host);
	if(host_info==NULL)
		return -1;
	memcpy((char*)&server.sin_addr, host_info->h_addr, host_info->h_length);
	server.sin_family = AF_INET;
	server.sin_port = htons(PORT);
	
	//Erzeuge den Socket
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock<0)
		return -1;
	
	//socket auf nonblocking in linux stellen
	#ifdef _WIN32
		unsigned long win32_blocking=1;
		ioctlsocket(sock, FIONBIO, &win32_blocking);
		
		//connect zu server
		if((connect(sock, (struct sockaddr*)&server, sizeof(server))==-1))
			if((WSAGetLastError()!=WSAEWOULDBLOCK) && (WSAGetLastError()!=WSAEINVAL))
				return -1;
		
		win32_blocking=0;
		ioctlsocket(sock, FIONBIO, &win32_blocking);
	#else
		old_fd = fcntl(sock, F_GETFL);
		new_fd = old_fd | O_NONBLOCK;
		fcntl(sock, F_SETFL, new_fd);
		
		//connect zu server
		if((connect(sock, (struct sockaddr*)&server, sizeof(server))==-1))
			if(errno!=EINPROGRESS)
				return -1;
		
		fcntl(sock, F_SETFL, old_fd);
	#endif
	
	memcpy(&global_server, &server, sizeof(server));
	global_socket=sock;
	return 0;
}

int wait_for_client_connect(void)
{
	int msg_num=_MSG_CONNECT_, protocol_version=_PROTOCOL_VERSION_(0,2);
	char msg_buffer[1024], message[1024];
	int size, con_ok=0, ret_val;
	struct timeval zeit;
	
	memset(msg_buffer, 0, 1024);
	memset(message, 0, 1024);
	
	#ifdef _WIN32
		FD_SET empfang;
		FD_SET connection;
		
		FD_ZERO(&connection);
		FD_SET(global_socket, &connection);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, NULL, NULL, &connection, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return -1;
		if(FD_ISSET(global_socket, &connection))
			return -1;				//connection failed
		
		FD_ZERO(&connection);
		FD_SET(global_socket, &connection);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, NULL, &connection, NULL, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return -1;
		if(FD_ISSET(global_socket, &connection))
			con_ok=1;				//connection succeeded
	#else	
		fd_set empfang;
		fd_set connection;
		
		FD_ZERO(&connection);
		FD_SET(global_socket, &connection);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, &connection, NULL, NULL, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return -1;
		if(FD_ISSET(global_socket, &connection))
		{
			if(recv(global_socket, message, 1024, MSG_PEEK)<1)
				return -1;			//connection failed
		}
		
		FD_ZERO(&connection);
		FD_SET(global_socket, &connection);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, NULL, &connection, NULL, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return -1;
		if(FD_ISSET(global_socket, &connection))
			con_ok=1;				//connection succeeded
	#endif
	
	if(con_ok)
	{
		if(!_protocol_send)
		{
			_protocol_send=1;
			//erstmal protocol version senden
			memcpy(msg_buffer, &msg_num, sizeof(msg_num));
			memcpy(&msg_buffer[sizeof(msg_num)], &protocol_version, sizeof(protocol_version));
			if(send(global_socket, msg_buffer, sizeof(msg_num)+sizeof(protocol_version), 0)<0)
				return -1;
		}
		
		FD_ZERO(&empfang);
		FD_SET(global_socket, &empfang);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, &empfang, NULL, NULL, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return -1;
		if(FD_ISSET(global_socket, &empfang))
		{
			size=recv(global_socket, message, sizeof(msg_num)+sizeof(protocol_version), 0);
			if(size<1)
				return -1;
			
			memcpy(&msg_num, message, sizeof(msg_num));
			memcpy(&protocol_version, message+sizeof(msg_num), sizeof(protocol_version));
			if((msg_num==_MSG_CONNECT_) && (protocol_version==_PROTOCOL_VERSION_(0,2)))
			{
				_protocol_send=0;
				#ifdef _WIN32
					closesocket(global_listen);
					global_listen=INVALID_SOCKET;
				#else
					close(global_listen);
					global_listen=-1;
				#endif
				return 1;
			}
			else
			{
				close_network();
				return -2;
			}
		}
	}
	return 0;
}

void close_network(void)
{
	#ifdef _WIN32
		closesocket(global_socket);
		closesocket(global_listen);
		global_socket=INVALID_SOCKET;
		global_listen=INVALID_SOCKET;
	#else
		close(global_socket);
		close(global_listen);
		global_socket=-1;
		global_listen=-1;
	#endif
	return;
}

int get_msg(int blocking)
{
	char message[1024];
	int size, msg_num, ret_val;
	
	if(blocking==_BLOCKING_)
	{
		size=recv(global_socket, message, sizeof(msg_num), 0);
		if(size<sizeof(msg_num))
			return _ERROR_;
		memcpy(&msg_num, message, sizeof(msg_num));
			return msg_num;
	}
	else if(blocking==_NON_BLOCKING_)
	{
		struct timeval zeit;
		#ifdef _WIN32
			FD_SET empfang;
		#else
			fd_set empfang;
		#endif
		FD_ZERO(&empfang);
		FD_SET(global_socket, &empfang);
		zeit.tv_sec=0;
		zeit.tv_usec=0;
		ret_val=select(global_socket+1, &empfang, NULL, NULL, &zeit);
		if(ret_val==EBADF || ret_val==EINTR || ret_val==EINVAL || ret_val==ENOMEM)
			return _ERROR_;
		if(FD_ISSET(global_socket, &empfang))
		{
			size=recv(global_socket, message, sizeof(msg_num), 0);
			if(size<sizeof(msg_num))
				return _ERROR_;
			
			memcpy(&msg_num, message, sizeof(msg_num));
			return msg_num;
		}
		else
			return _NO_DATA_;
	}
	else
		return -128;
}

int get_update_data(int *feld, int *x, int *y)
{
	int len=sizeof(*x)+sizeof(*y)+(1160/8);		//x, y, feld[1160]
	int size, msg_pos=0, feld_pos, bit_pos;
	char message[1024];
	
	//warten, bis komplette message empfangen wurde (quasi blocking, aber ohne cpu-freigabe)
	do
	{
		size=recv(global_socket, message, len-1, MSG_PEEK);		//156 mit msg_num
		if(size<1)
			return _ERROR_;
		else
			printf("len=%d, size: %d...\n", len, size); fflush(stdout);
	} while(size<len-1);
	
	size=recv(global_socket, message,len-1, 0);
	if(size<len-1)
		return _ERROR_;
	
	//x-, y-koordinate einlesen
	memcpy(x, message+msg_pos, sizeof(*x)); msg_pos+=sizeof(*x);
	memcpy(y, message+msg_pos, sizeof(*y)); msg_pos+=sizeof(*y);
	size=msg_pos;
	//feld einlesen
	for(feld_pos=0;feld_pos<1160;feld_pos++)
	{
		msg_pos=size+((int)(feld_pos/8));
		bit_pos=feld_pos%8;
		if(message[msg_pos] & (1<<bit_pos))
			feld[feld_pos]=8;
		else
			feld[feld_pos]=0;
	}
	
	return 0;
}

int write_msg(int msg_num)
{
	if(send(global_socket, &msg_num, sizeof(msg_num), 0)<sizeof(msg_num))
		return _ERROR_;
	return 0;
}

int write_update_msg(int *feld, int x, int y)
{
	char message[1024];
	int msg_num=_MSG_UPDATE_;
	int msg_pos=0, feld_pos, bit_pos;
	char byte=0, size;
	
	//msg_num senden
	memcpy(message+msg_pos, &msg_num, sizeof(msg_num)); msg_pos+=sizeof(msg_num);
	//x- und y-koordinate senden
	memcpy(message+msg_pos, &x, sizeof(x)); msg_pos+=sizeof(x);
	memcpy(message+msg_pos, &y, sizeof(y)); msg_pos+=sizeof(y);
	//feld senden
	size=msg_pos;
	for(feld_pos=0;feld_pos<1160;feld_pos++)
	{
		msg_pos=size+((int)(feld_pos/8));
		bit_pos=feld_pos%8;
		byte|=((int)(feld[feld_pos]/8))<<bit_pos;
		if(bit_pos==7)
		{
			message[msg_pos]=byte;
			byte=0;
		}
	}
	//daten senden
	if(send(global_socket, message, msg_pos, 0)<sizeof(msg_pos))
		return _ERROR_;
	return 0;
}

